#!/usr/bin/python3
#  coding: utf-8
#
#       setPress_getDisp.py
#
#       OpenFOAMの圧力pをFrontISTRのcntファイルにセットし、
#       変位の点群ファイルを作成する。
# 　　　変位の点群ファイルからOFのpointにmappingした結果fileを
#       作成する。（OF側がこのfileを読み込む）
#
#   23/01/24    新規作成
#      04/05    couplingの判断をbeginning後に変更
#      05/17    isCoupling:fsi,chtssのcouplingを修正
#


import sys
import os
import time
import glob
import mmap
import mappingFromPointCloud as mapping
import pyCoupling

waitCount = 5000
waitTime = 0.01

commonSize = 10000



#
#  getMappedPointLocs
#---------------------
def getMappedPointLocs():
    """ mappingするpointの座標を取得する"""
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mapPointFaceDict = {}
    mapPointDict = {}
    for line in lines:
        words = line.split()
        if len(words) > 4:
            nodeNo = int(words[0])
            pointLoc = list(map(float, words[1:4]))
            faceNodes = list(map(int, words[4:]))
            mapPointFaceDict[nodeNo] = [pointLoc, faceNodes]
            mapPointDict[nodeNo] = pointLoc
    return mapPointFaceDict, mapPointDict

#
# createPtcFile
#---------------
def createPtcFile(pressFile):
    """ 圧力の点群file作成"""
    global mmList
    #faceCenterLoc読み込み
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    faceCenterFile = dataDir + "/" + patchName + "_faceCenterLocs_" + procNo
    f = open(faceCenterFile); lines = f.readlines(); f.close()
    #facePressを読み込み
    fp = open(pressFile); pLines = fp.readlines(); fp.close()
    newLines = []
    for i in range(len(lines)):
        newLine = lines[i][:-1] + " " + pLines[i]
        newLines.append(newLine)
    newLines.append("#end\n")
    #streamにセットし、送信
    cont = "".join(newLines)
    title = "pressurePtcData"
    streamToParent.put(title, cont)
    #圧力のptcファイルを保存
    name = patchName + "_faceCenterPress_" + beforeTime + "_" + procNo + ".particles"
    ptcFileName = solidCaseDir + "/" + name
    f = open(ptcFileName, "w"); f.writelines(newLines); f.close()
    return

#
#  saveMappingDisp
#------------------
def saveMappingDisp(mappingDict):
    """ mappingDispの結果を保存。このファイルをOF側が読み込む"""
    saveDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = saveDir + "/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    #保存のfilename
    name = patchName + "_disp_" + beforeTime + "_" + procNo
    dispFile = saveDir + "/" + name
    f = open(dispFile, "w")
    for line in lines:
        words = line.split()
        nodeNo = int(words[0])
        disp = mappingDict[nodeNo]
        line = " ".join(list(map(str, disp))) + "\n"
        f.write(line)
    f.close()
    if procNo == "0":
        print("displacement value was created for OpenFOAM at " + timeName + ".")
    return

#
#  getLocDisplacement
#---------------------
def getLocDisplacement():
    """ 変位の点群ファイルStreamから読み込み"""
    #straemから読み込み
    #lines = getStreamLines("displacement")
    lines = streamToParent.getLines("displacement")
    locDisp = []
    for line in lines:
        if line[0] != "#":
            words = line.split()
            loc = list(map(float, words[:3]))
            disp = list(map(float, words[3:]))
            locDisp.append([loc, disp])
    return locDisp

#
#  isMappingTriFile
def isMappingTriFile():
    """ mapping用の3点を取得したfileが存在するかチェック"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    return os.path.exists(fileName)

#
#  createMappingTriFile
def createMappingTriFile(mappingTriDict):
    """ mapping用の3点の座標をfileに保存する"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    lines = []
    for key in mappingTriDict.keys():
        faceNo = key
        word = str(faceNo)
        triLocs = mappingTriDict[key]
        loc0, loc1, loc2 = triLocs
        word0 = " ".join(list(map(str, loc0)))
        word1 = " ".join(list(map(str, loc1)))
        word2 = " ".join(list(map(str, loc2)))
        line = " ".join([word, word0, word1, word2]) + "\n"
        lines.append(line)
    f = open(fileName, "w"); f.writelines(lines); f.close()

#
#  readMappingTriFile
def readMappingTriFile():
    """ 保存したmapping用の3点の座標を読み込む"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    mappingTriDict = {}
    for line in lines:
        words = line.split()
        faceNo = int(words[0])
        key = faceNo
        loc0 = list(map(float, words[1:4]))
        loc1 = list(map(float, words[4:7]))
        loc2 = list(map(float, words[7:10]))
        mappingTriDict[key] = [loc0, loc1, loc2]
    return mappingTriDict

#---------------------
#  setPress_getDisp
#---------------------
def setPress_getDisp():
    """ pressをmappingしてfistrの結果を待ち、変位を取得する"""
    mesh = mapping.fistrMesh(solidCaseDir)
    #ptcDataの読み込み
    ptcDir = fluidCaseDir + "/coupling_FrontISTR/data"
    pressFile = ptcDir + "/" + patchName + "_facePress_" + beforeTime + "_" + procNo
    #圧力の点群ファイルを作成し、送信 
    createPtcFile(pressFile)
    #mappingするpointの座標を予め取得しておく（待ち時間を利用）
    mapPointFaceDict, mapPointDict = getMappedPointLocs()
    #変位の点群dataをstreamから読み込み
    locDisp = getLocDisplacement()
    #mappingTriFile(mapping用の3点の座標を保存)があるか？
    if isMappingTriFile() == False:    
        #存在しない場合、mappingTriDictを取得
        mapStat, mappingTriDict = mesh.getMappingTriLocs(locDisp, mapPointFaceDict, mapPointDict)
        if mapStat != "OK":
            return "error in displacement mapping."
        #mappingTriFileを作成
        createMappingTriFile(mappingTriDict)
    else:
        #fileを読み込む
        mappingTriDict = readMappingTriFile()
    #変位をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locDisp, mapPointFaceDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingDisp(mappingDict)
    return ""

#---------------
#  getChtssDisp
#---------------
def getChtssDisp():
    """ chtss変位を取得する"""
    mesh = mapping.fistrMesh(solidCaseDir)
    #ptcDataの読み込み
    ptcDir = fluidCaseDir + "/coupling_FrontISTR/data"
    #mappingするpointの座標を取得
    mapPointFaceDict, mapPointDict = getMappedPointLocs()
    #変位の点群dataをstreamから読み込み
    locDisp = getLocDisplacement()
    #mappingTriFile(mapping用の3点の座標を保存)があるか？
    if isMappingTriFile() == False:    
        #存在しない場合、mappingTriDictを取得
        mapStat, mappingTriDict = mesh.getMappingTriLocs(locDisp, mapPointFaceDict, mapPointDict)
        if mapStat != "OK":
            return "error in displacement mapping."
        #mappingTriFileを作成
        createMappingTriFile(mappingTriDict)
    else:
        #fileを読み込む
        mappingTriDict = readMappingTriFile()
    #変位をmapping
    mappingDict = mesh.setMappingDataFromTriLocs(locDisp, mapPointFaceDict, mappingTriDict)
    #mapping結果を保存(このfileをOpenFOAM側が読み込む)
    saveMappingDisp(mappingDict)
    return ""

#
#  beginning
#------------
def beginning():
    """ beforeTime, currTimeをファイルに保存"""
    #processor0のみ
    if procNo == "0":
        #時間を書き込み
        cont = pyCoupling.createTimeCont(beforeTime, timeName)
        # lines =  ["beforeTime " + beforeTime + "\n"]
        # lines += ["currTime   " + timeName + "\n"]
        # lines += ["coupling   yesno\n"]
        # lines += ["#end\n"]
        # cont = "".join(lines)
        mmTimes.seek(0)
        mmTimes.write(cont.encode())
        mmTimes.flush()
    #couplingFileに反映されるまで待つ
    waitUntilCouplingWritten(beforeTime)
    return

#
#  waitUntilCouplingWritten
def waitUntilCouplingWritten(beforeTime):
    """ couplingStepにbeforeTimeが反映されるまで待つ"""
    count = waitCount
    sleepTime = waitTime
    while True:
        mmCoupling.seek(0)
        cont = mmCoupling.read().decode()
        lines = cont.split("\n")
        if len(lines) >= 3:
            bTime = lines[0].split()[1]
            if float(bTime) == float(beforeTime):
                #beforeTimeが反映された場合戻る
                break
        count -= 1
        if count < 0:
            break
        time.sleep(sleepTime)
    return

#
#  createMmapFile
#-----------------
def createMmapFile(fileName):
    """ mmapFileを作成する"""
    f = open(fileName, "r+b"); mmFile = mmap.mmap(f.fileno(), 0); f.close()
    return mmFile

#
#  createInitialDispFile
#------------------------
def createInitialDispFile():
    """ 初期の変位dataを作成し、セットする"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_mapPointFaceNodes_" + procNo
    fileName = dataDir + "/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    #保存のfileName
    name = patchName + "_disp_" + beforeTime + "_" + procNo
    dispFile = dataDir + "/" + name
    dataLine = "0.0 0.0 0.0\n"
    f = open(dispFile, "w"); 
    for i in range(len(lines)):
        f.write(dataLine)
    f.close()

#
#  createInitialPressFile
#-------------------------
def createInitialPressFile():
    """ 初期の圧力dataを作成し、セットする。"""
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    name = patchName + "_faceCenterLocs_" + procNo
    fileName = dataDir + "/" + name
    f = open(fileName); lines = f.readlines(); f.close()
    #保存のfileName
    name = patchName + "_facePress_" + beforeTime + "_" + procNo
    pressFile = dataDir + "/" + name
    dataLine = "0.0\n"
    f = open(pressFile, "w"); 
    for i in range(len(lines)):
        f.write(dataLine)
    f.close()

#
#  deleteMappingTriFile
#-----------------------
def deleteMappingTriFile():
    """ 初回起動時に、mappingTriFileを削除する"""
    name = "mappingTriFile_" + procNo
    fileName = fluidCaseDir + "/coupling_FrontISTR/data/" + name
    if os.path.exists(fileName) == True:
        os.remove(fileName)

#
#  getOlderTime
def getOlderTime(bTime):
    """ bTimeの1個古いtimeを取得する"""
    mmTimeList.seek(0)
    conts = mmTimeList.read().decode()
    strTimes = conts.split("\n")
    if bTime in strTimes:
        id = strTimes.index(bTime)
        oldTime = strTimes[id+1]
        try:
            float(oldTime)
        except:
            oldTime = "-1"
    else:
        oldTime = "-1"
    return oldTime

#
#  copyMapPointDispFile
#-----------------------
def copyMapPointDispFile():
    """ mappingDispのfileをrenameする"""
    #計算結果があるか？
    newName = patchName + "_disp_" + beforeTime + "_" + procNo
    newFileName = fluidCaseDir + "/coupling_FrontISTR/data/" + newName
    if os.path.exists(newFileName) == True:
        return
    #beforeTimeの直前のtimeを取得
    oldTime = getOlderTime(beforeTime)
    oldName = patchName + "_disp_" + oldTime + "_" + procNo
    oldFileName = fluidCaseDir + "/coupling_FrontISTR/data/" + oldName
    #oldFileNameが存在するか？
    if os.path.exists(oldFileName) == False:
        #存在しない場合、最新のfileを取得
        name = patchName + "_disp_*_" + procNo
        pattern = fluidCaseDir + "/coupling_FrontISTR/data/" + name
        files = glob.glob(pattern)
        fileTimes = []
        for fileName in files:
            ft = fileName.split("_")[-2]
            fileTimes.append([float(ft), ft, fileName])
        fileTimes.sort()
        oldName = patchName + "_disp_" + fileTimes[-1][1] + "_" + procNo
        oldFileName = fluidCaseDir + "/coupling_FrontISTR/data/" + oldName
    os.rename(oldFileName, newFileName)

#
#  isCouplingTime
#-----------------
def isCouplingTime():
    """ couplingTime（連成させる時間）かどうか確認
    skipSeps=1の場合、次はcouplingになる。"""
    coupling = False
    couplingTS = False
    if int(nStepsOF) == 1:
        #nStepOF=1の場合は、毎回連成
        coupling = True
        return coupling, couplingTS
    deltaT = str(float(timeName) - float(beforeTime))
    if pyCoupling.isWriteInterval(timeName, writeInterval, deltaT) == True:
        #writeIntervalでは連成する
        coupling = True
        return coupling, couplingTS

    mmCoupling.seek(0)
    conts = mmCoupling.read().decode()
    lines = conts.split("\n")
    if len(lines) < 3:
        #計算初回は連成
        coupling = True
        return coupling, couplingTS
    bTime = lines[0].split()[1]
    cTime = lines[1].split()[1]
    skipSteps = int(lines[2].split()[1])
    skipStepsTemp = int(lines[3].split()[1])
    skipStepsTS = int(lines[4].split()[1])
    if float(bTime) == 0:
        #初回は、連成
        coupling = True
        return coupling, couplingTS
    if float(beforeTime) == float(startTime):
        coupling = True
        return coupling, couplingTS
    #fsi, chtssのcouplingを確認
    if skipSteps == 0:
        coupling = True
    if skipStepsTS == 0:
        couplingTS = True
    return coupling, couplingTS


if __name__ == "__main__":
    import gettext
    gettext.install("app") # replace with the appropriate catalog name
    _ = gettext.gettext

    timeName = sys.argv[1]      #currTime
    beforeTime = sys.argv[2]    #前回のtime
    patchName = sys.argv[3]     #patchName
    procNo = sys.argv[4]        #procNo
    nProcs = sys.argv[5]        #nProcs
    fluidCaseDir = os.getcwd()
    couplingData = pyCoupling.couplingData(fluidCaseDir)
    couplingDict = couplingData.read()
    startTime = couplingDict["startTime"]
    writeInterval = couplingDict["writeInterval"]
    nStepsOF = couplingDict["nStepsOF"]
    streamBuffer = couplingDict["streamBuffer"]
    maxWaitTime = couplingDict["maxWaitTime"]
    solidCaseDir = couplingDict["solidCaseDir"]
    sgrpName = couplingDict["mappingSgrp"]
    mapCoeff = couplingDict["mappingCoeff"]
    solidAnaComm = couplingDict["solidAnaComm"]
    nSteps = couplingDict["nSteps"]
    waitCount = int(maxWaitTime)
    #共有メモリの設定
    commonSize = int(streamBuffer)
    dataDir = fluidCaseDir + "/coupling_FrontISTR/data"
    mmCommFile = createMmapFile(dataDir + "/common_" + procNo)
    mmTimes = createMmapFile(dataDir + "/times")
    mmCoupling = createMmapFile(dataDir + "/couplingStep")
    mmTimeList = createMmapFile(dataDir + "/timeList")
    #streamを定義
    streamToParent = pyCoupling.stream(mmCommFile, commonSize,
                                       waitCount, waitTime )
    beginning()
    #連成させる時間？
    isCouple = isCouplingTime()
    if isCouple[0] == False:
        #連成させない時間の場合、前回の結果をコピーして終了。
        copyMapPointDispFile()
        if isCouple[1] == False:
            exit()
        #chtssのみで連成させる
        solidCaseDir = os.path.abspath(solidCaseDir)
        stat = getChtssDisp()
        if stat != "":
            raise NameError(stat)
        exit()    
    #計算開始時？（初期化の有無）
    elif (float(beforeTime) == float(startTime) 
        and float(startTime) == 0.0):
        #初回起動時の設定
        deleteMappingTriFile()
        createInitialPressFile()
    #途中から再開した場合（latestTime）
    elif (float(beforeTime) == float(startTime)
        and int(nStepsOF) > 1):
        #結果fileがあるか？
        #if isResultFile() == True:
        #    exit()
        pass
    #連続計算
    solidCaseDir = os.path.abspath(solidCaseDir)
    stat = setPress_getDisp()
    if stat != "":
        raise NameError(stat)
