#!/usr/bin/python3
#  coding: utf-8
#
#           pyCoupling.py
#
#       fsi, cht, chtssに応じたcnt, mshファイルを作成する
#
#   23/04/21    新規作成
#      05/28    couplingData.write:修正。
#

import os
import shutil
import time
import mmap
import pyFistr_ptc as pyFistr


waitCount = 5000
waitTime = 0.01
commonSize = 10000

#-----------------
#  stream class
#-----------------
class stream:
    """ 双方向通信"""
    def __init__(self, mmComm, commonSize, waitCount, waitTime):
        self.mmComm = mmComm            #mmapFile
        self.commonSize = commonSize    #bufferサイズ
        self.waitCount = waitCount      #count数
        self.waitTime = waitTime        #count当たりの待ち時間

    #
    #  put
    #--------
    def put(self, title, text):
        """ データ送信"""
        size = self.commonSize
        mmCommFile = self.mmComm
        myPid = os.getpid()
        header = "PID:" + str(myPid) + " title:" + title
        cont = header + "\01" + text + "\04"
        cont = cont.encode()
        dataList = []
        for i in range(0, len(cont), size):
            data = cont[i:i+size]
            dataList.append(data)
        #textをstreamにセットする
        for i in range(len(dataList)):
            if i == 0:
                self.waitUntilBufferInitialSet()
            else:
                self.waitUntilBufferEmpty()
            data = dataList[i]
            mmCommFile.seek(0)
            mmCommFile.write(data)
            mmCommFile.flush()
        mmCommFile.seek(commonSize-1)
        mmCommFile.write(b"\04")        #最後に書き込み
        mmCommFile.flush()
        self.waitUntilBufferEmpty()     #相手が読み込むまで待つ

    #
    #  waitUntilBufferInitialSet
    def waitUntilBufferInitialSet(self):
        """ bufferがemptyの場合、直ぐにPIDをセットし、bufferを使用中にする"""
        count = self.waitCount
        sleepTime = self.waitTime
        mmCommFile = self.mmComm
        bPid = str(os.getpid()).encode()
        while True:
            mmCommFile.flush()
            if mmCommFile[:2] == b"\00\00" and mmCommFile[-1:] == b"\00":
                mmCommFile.write(bPid)
                mmCommFile.flush()
                mmCommFile.seek(0)
                break
            count -= -1
            if count < 0:
                break
            time.sleep(sleepTime)
        return

    #
    #  waitUntilBufferEmpty
    def waitUntilBufferEmpty(self):
        """ 送信可能になるまで待つ。bufferが空になるまで"""
        count = self.waitCount
        sleepTime = self.waitTime
        mmCommFile = self.mmComm
        while True:
            if mmCommFile[:2] == b"\00\00" and mmCommFile[-1:] == b"\00":
                break
            count -= -1
            if count < 0:
                break
            time.sleep(sleepTime)
        return

    #
    #  waitUntilBufferSet
    def waitUntilBufferSet(self):
        """ streamに文字がセットされるまで待つ"""
        count = self.waitCount
        sleepTime = self.waitTime
        mmCommFile = self.mmComm
        while True:
            mmCommFile.flush()
            if mmCommFile[:2] != b"\00\00" and mmCommFile[-1:] != b"\00":
                break
            count -= 1
            if count < 0:
                break
            time.sleep(sleepTime)
        return  

    #
    #  get
    #-------
    def get(self, title):
        """ データ受信"""
        mmCommFile = self.mmComm
        commonSize = self.commonSize
        #dataがセットされるまで待つ
        self.waitUntilBufferSet()
        #buffer読み込み
        mmCommFile.seek(0)
        conts = mmCommFile.read().decode()
        #bufferクリア
        mmCommFile.seek(0)
        mmCommFile.write(b"\00\00")
        mmCommFile.seek(commonSize-1)
        mmCommFile.write(b"\00")
        mmCommFile.flush()
        #streamのheaderを取得
        ns = conts.find("\01")
        header = conts[:ns]
        #header内のtitleを取得
        nt = header.find("title:")
        title = header[nt + len("title:"):]
        #streamのtextを取得
        text = ""
        cont = conts[ns+1:]
        ne = cont.find("\04")
        if ne >= 0:
            text = cont[:ne]
            return title, text
        text += cont
        while True:
            self.waitUntilBufferSet()
            #読み込み
            mmCommFile.seek(0)
            cont = mmCommFile.read().decode()
            #bufferクリア
            mmCommFile.seek(0)
            mmCommFile.write(b"\00\00")
            mmCommFile.seek(commonSize-1)
            mmCommFile.write(b"\00")
            mmCommFile.flush()
            #読み込み内容を処理
            ne = cont.find("\04")
            if ne >= 0:
                text += cont[:ne]
                break
            text += cont
        return title, text

    #
    #  getLines
    #------------
    def getLines(self, title):
        """ 受信データをlinesで取得"""
        _title, text = self.get(title)
        if text[-1] == "\n":
            text = text[:-1]
        lines = text.split("\n")
        lines = list(map(lambda x: x+"\n", lines))
        return lines


#--------------------
#  cntMshFiles class
#--------------------
class cntMshFiles:
    """ cnt, msh, datファイルをチェックし、無ければ作成する。"""

    def __init__(self, solidCaseDir):
        self.solidCaseDir = solidCaseDir

    #
    #  checkCntMshFiles
    def checkCntMshFiles(self):
        """ 標準のcnt、mshファイルかどうかチェックし、
        標準でない場合、標準に戻す。"""
        changeDict = {}
        nameDict = pyFistr.getCntMshFileNameDict(self.solidCaseDir)
        if nameDict["cnt"] != "FistrModel.cnt":
            orgName = self.solidCaseDir + os.sep + nameDict["cnt"]
            newName = self.solidCaseDir + os.sep + "FistrModel.cnt"
            os.rename(orgName, newName)
            changeDict["cnt"] = "FistrModel.cnt"
        if nameDict["msh"] != "FistrModel.msh":
            orgName = self.solidCaseDir + os.sep + nameDict["msh"]
            newName = self.solidCaseDir + os.sep + "FistrModel.msh"
            os.rename(orgName, newName)
            changeDict["msh"] = "FistrModel.msh"
        if len(changeDict) > 0:
            pyFistr.setCntMshFileNameDict(self.solidCaseDir, changeDict)

    #
    #  checkFsiFiles
    def checkFsiFiles(self):
        """ hecmw_ctrl_fsi.datが無い場合、作成する。
        既に存在すれば、何もしない"""
        solidCaseDir = self.solidCaseDir
        datFsiFile = solidCaseDir + os.sep + "hecmw_ctrl_fsi.dat"
        cntFsiFile =  solidCaseDir + os.sep + "FistrModel_fsi.cnt"
        if os.path.exists(datFsiFile) and os.path.exists(cntFsiFile):
            return
        #fsi専用のcntを作成
        FistrModel_cnt, FistrModel_msh = pyFistr.cntMshFileName(solidCaseDir)
        copyFile = solidCaseDir + os.sep + FistrModel_cnt
        pasteFile = solidCaseDir + os.sep + "FistrModel_fsi.cnt"
        if os.path.exists(pasteFile) == False:
            shutil.copy(copyFile, pasteFile)
        #fsi専用のdatを作成
        nameDict = {
            "msh":      "FistrModel.msh",
            "cnt":      "FistrModel_fsi.cnt",
            "res":      "FistrModel.res",
            "vis":      "FistrModel.vis",
            "restart":  "FistrModel.restart"
            }
        pyFistr.setCntMshFileNameDict(solidCaseDir, nameDict)
        copyFile = solidCaseDir + os.sep + "hecmw_ctrl.dat"
        if os.path.exists(datFsiFile) == False:
            shutil.copy(copyFile, datFsiFile)

    #
    #  checkChtFiles
    def checkChtFiles(self):
        """ hecmw_ctrl_cht.datが無い場合、作成する。
        既に存在すれば、何もしない"""
        solidCaseDir = self.solidCaseDir
        datChtFile = solidCaseDir + os.sep + "hecmw_ctrl_cht.dat"
        cntChtFile = solidCaseDir + os.sep + "FistrModel_cht.cnt" 
        if os.path.exists(datChtFile) and os.path.exists(cntChtFile):
            return
        #cht専用のcntを作成
        FistrModel_cnt, FistrModel_msh = pyFistr.cntMshFileName(solidCaseDir)
        copyFile = solidCaseDir + os.sep + FistrModel_cnt
        pasteFile = solidCaseDir + os.sep + "FistrModel_cht.cnt"
        if os.path.exists(pasteFile) == False:
            shutil.copy(copyFile, pasteFile)
        #cht専用のmshを作成
        copyFile = solidCaseDir + os.sep + FistrModel_msh
        pasteFile = solidCaseDir + os.sep + "FistrModel_cht.msh"
        if os.path.exists(pasteFile) == False:
            shutil.copy(copyFile, pasteFile)
        #cht専用のdatを作成
        nameDict = {
            "msh":      "FistrModel_cht.msh",
            "cnt":      "FistrModel_cht.cnt",
            "res":      "FistrModel_temp.res",
            "vis":      "FistrModel_cht.vis",
            "restart":  "FistrModel_cht.restart"
            }
        pyFistr.setCntMshFileNameDict(solidCaseDir, nameDict)
        copyFile = solidCaseDir + os.sep + "hecmw_ctrl.dat"
        if os.path.exists(datChtFile) == False:
            shutil.copy(copyFile, datChtFile)

    #
    #  checkChtssFiles
    def checkChtssFiles(self):
        """ hecmw_ctrl_chtss.datが無い場合、作成する。
        既に存在すれば、何もしない"""
        solidCaseDir = self.solidCaseDir
        datChtssFile = solidCaseDir + os.sep + "hecmw_ctrl_chtss.dat"
        cntChtssFile = solidCaseDir + os.sep + "FistrModel_chtss.cnt"
        if os.path.exists(datChtssFile) and os.path.exists(cntChtssFile):
            return
        #chtss専用のcntを作成
        FistrModel_cnt, FistrModel_msh = pyFistr.cntMshFileName(solidCaseDir)
        copyFile = solidCaseDir + os.sep + FistrModel_cnt
        pasteFile = solidCaseDir + os.sep + "FistrModel_chtss.cnt"
        if os.path.exists(pasteFile) == False:
            shutil.copy(copyFile, pasteFile)
        #chtss専用のdatを作成
        nameDict = {
            "msh":      "FistrModel.msh",
            "cnt":      "FistrModel_chtss.cnt",
            "res":      "FistrModel.res",
            "vis":      "FistrModel_chtss.vis",
            "restart":  "FistrModel_chtss.restart"
            }
        pyFistr.setCntMshFileNameDict(solidCaseDir, nameDict)
        copyFile = solidCaseDir + os.sep + "hecmw_ctrl.dat"
        if os.path.exists(datChtssFile) == False:
            shutil.copy(copyFile, datChtssFile)

#----------------------
#  couplingData class
#----------------------
class couplingData:
    """ couplingDataのread、write"""

    def __init__(self, fluidCaseDir):
        self.fluidCaseDir = fluidCaseDir
        self.dataFile = fluidCaseDir + "/coupling_FrontISTR/couplingData"

    #
    #  readMmapLines
    def readMmapLines(self):
        """ fileObjectの共有メモリからlineを読込、返す。"""
        f = open(self.dataFile, "r+b")
        mm = mmap.mmap(f.fileno(), 0)
        f.close()
        mm.seek(0)
        conts = mm.read().decode()
        lines = conts.split("\n")
        mm.close()
        return lines

    #
    #  read
    #---------
    def read(self):
        """ couplingDataを辞書形式で読み込み、返す。"""
        #値の初期化
        dataDict = {
            "field": "",                #OpenFOAM関連
            "pointField": "",
            "mappingPatch": "",
            "mappingCoeff": "1.0",
            "mappingShift": "0.0",
            "nStepsOF": "",
            "nProcs": "",
            "startTime": "",
            "deltaT": "",
            "writeInterval": "",

            "solidAnaComm": "fistr1",   #FrontISTR関連
            "solidCaseDir": "",
            "mappingSgrp": "",
            "nSteps": "",
            "nThreads": "",

            "streamBuffer": "",         #その他
            "maxWaitTime": "",
            "fsi": "",
            "cht": "",
            "chtss": ""
            }
        #共有メモリから読み込む
        lines = self.readMmapLines()
        for line in lines:
            words = line.split()
            if len(words) >= 2:
                if words[0][0] != "#":
                    dataDict[words[0]] = words[1]
        return dataDict
    
    #
    #  write
    #---------
    def write(self, dataDict):
        """ couplingDataを書き込む"""
        fileName = self.dataFile
        f = open(fileName); lines = f.readlines(); f.close()
        addLines = []
        for key in dataDict.keys():
            cont = dataDict[key]
            flag = 0
            for i in range(len(lines)):
                line = lines[i]
                words = line.split()
                if len(words) >= 1:
                    if words[0] == key:
                        if len(words) >= 2:
                            words[1] = cont
                            word = ("%-15s" % words[0] + " ")
                            lines[i] = word + " ".join(words[1:]) + "\n"
                            flag = 1
                            break
                        else:
                            words += [cont]
                            word = ("%-15s" % words[0] + " ")
                            lines[i] = word + " ".join(words[1:]) + "\n"
                            flag = 1
                            break
            if flag == 0:
                word = ("%-15s" % key + " ")
                line = word + cont + "\n"
                addLines.append(line)
        lines += addLines
        f = open(fileName, "w"); f.writelines(lines); f.close()


#
#  isWriteInterval
#------------------
def isWriteInterval(currTime, writeInterval, deltaT):
    """ currTimeがwriteIntervalのタイミングかどうかチェックする"""
    dt = float(deltaT)
    curr = float(currTime)
    wInt = float(writeInterval)
    diffR = curr / wInt - int(curr/ wInt + 0.5)
    diff = diffR * wInt
    if abs(diff) < dt/2.0:
        return True
    else:
        return False
    
#
#  createTimeCont
#--------------------
def createTimeCont(beforeTime, currTime):
    """ 共有メモリmmTimesの内容を作成して返す"""
    lines =  ["beforeTime   " + beforeTime + "\n"]
    lines += ["currTime     " + currTime + "\n"]
    lines += ["coupling     yesno\n"]
    lines += ["couplingTemp yesno\n"]
    lines += ["couplingTS   yesno\n"]
    lines += ["#end\n"]
    cont = "".join(lines)
    return cont
